"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of betbot.

betbot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

betbot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with betbot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import requests
from math import sqrt
from bs4 import BeautifulSoup
from typing import List, Tuple
from betbot.api.Bet import Bet
from betbot.api.Match import Match
from betbot.prediction.Predictor import Predictor


class TipicoOddsPredictor(Predictor):
    """
    Class that determinalistically predicts matches based on Tipico quotes
    """

    @classmethod
    def name(cls) -> str:
        """
        :return: The name of the predictor
        """
        return "tipico-odds"

    def predict(self, matches: List[Match]) -> List[Bet]:
        """
        Performs the prediction
        :param matches: The matches to predict
        :return: The predictions as Bet objects
        """
        tipico = self.load_tipico_data(matches[0].matchday)
        bets = []

        for home_team, away_team, home_odds, draw_odds, away_odds in tipico:
            for match in matches:

                if home_team == match.home_team \
                        and away_team == match.away_team:
                    bet = self.generate_bet(
                        match.id, home_odds, draw_odds, away_odds
                    )
                    bets.append(bet)

        return bets

    # noinspection PyMethodMayBeStatic
    def load_tipico_data(self, matchday: int) \
            -> List[Tuple[str, str, float, float, float]]:
        """
        Loads tipico quote data for the 2020/21 bundesliga season
        :param matchday: The matchday for which to retrieve the data
        :return: The quote data as a list of tuples of the form:
                    home team name, away team name,
                    home win odds, draw odds, away win odds
        """
        tipico_data = []
        url = f"https://s5.sir.sportradar.com/tipico/de/1/season" \
              f"/77189/fixtures/round/21-{matchday}"
        req = requests.get(url)
        soup = BeautifulSoup(req.text, "html.parser")
        matches = soup.findAll("tr", {"class": "cursor-pointer"})
        for match in matches:
            names = match.findAll("div", {"class": "wrap"})
            odds = match.findAll("button", {"class": "odds-normal-content"})
            home = names[2].text
            away = names[5].text
            home_odds = float(odds[0].text)
            draw_odds = float(odds[1].text)
            away_odds = float(odds[2].text)

            translator = {
                "ARB": "BIE",
                "KOE": "FCK",
                "UNI": "SCU",
                "WOB": "VFL"
            }
            home = translator.get(home, home)
            away = translator.get(away, away)

            tipico_data.append((home, away, home_odds, draw_odds, away_odds))
        return tipico_data

    # noinspection PyMethodMayBeStatic
    def generate_bet(
            self,
            match_id: int,
            home_odds: float,
            draw_odds: float,
            away_odds: float
    ) -> Bet:
        """
        Generates a Bet based on the quote data
        :param match_id: The ID of the match to vote on
        :param home_odds: The odds that the home team wins
        :param draw_odds: The odds that the match ends in a draw
        :param away_odds: The odds that the away team wins
        :return: The generated bet
        """
        winner = int(sqrt(abs(home_odds - away_odds) * 2))
        loser = int(min(home_odds, away_odds) - 1)

        if home_odds < away_odds:
            home_score, away_score = winner, loser
        else:
            home_score, away_score = loser, winner

        if draw_odds > home_odds and draw_odds > away_odds:
            home_score += 1
            away_score += 1

        return Bet(match_id, home_score, away_score)



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of betbot.

betbot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

betbot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with betbot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Dict, Any


class Match:
    """
    Class that encapsulates information about a Match
    """

    def __init__(
            self,
            _id: int,
            matchday: int,
            home_team: str,
            away_team: str
    ):
        """
        Initializes the Match
        :param _id: The ID of the match
        :param matchday: The matchday of the match
        :param home_team: The name of the home team
        :param away_team: The name of the away team
        """
        self.id = _id
        self.matchday = matchday
        self.home_team = home_team
        self.away_team = away_team

    @classmethod
    def from_json(cls, json_data: Dict[str, Any]):
        """
        Generates a Match object from JSON data
        :param json_data: The JSON data
        :return: The generated Match
        """
        return cls(
            json_data["id"],
            json_data["matchday"],
            json_data["home_team"]["abbreviation"],
            json_data["away_team"]["abbreviation"]
        )



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of betbot.

betbot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

betbot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with betbot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import os
import json
import requests
import logging
from typing import List, Dict, Optional, Any
from base64 import b64encode
from betbot.api.Bet import Bet
from betbot.api.Match import Match


class ApiConnection:
    """
    Class that handles API calls to the bundesliga-tippspiel instance
    """

    def __init__(self, username: str, password: str, url: str):
        """
        Intitializes the API connection
        :param username: The username on bundesliga-tippspiel
        :param password: The password for that user
        :param url: The base url to the bundesliga-tippspiel instance
        """
        self.logger = logging.getLogger(__name__)
        self.username = username
        self.password = password
        self.url = os.path.join(url, "api/v2") + "/"
        self.api_key = ""
        self.login()

    @property
    def auth_headers(self) -> Dict[str, str]:
        """
        :return: Authorization headers for API calls
        """
        return {"Authorization": "Basic " + self.api_key}

    def login(self) -> bool:
        """
        Retrieves an API Key using a username and password if no key
        has been retrieved yet or if the existing key has expired
        :return: True if the login was successfulle
        """
        if not self.authorized():
            self.logger.info("Requesting new API Key")
            creds = {"username": self.username, "password": self.password}
            data = self.execute_api_call("key", "POST", False, creds)

            if data["status"] != "ok":
                self.logger.warning("Login attempt failed")
                return False
            else:
                api_key = data["data"]["api_key"]
                encoded = b64encode(api_key.encode("utf-8"))
                self.api_key = encoded.decode("utf-8")
                return True
        else:
            return True

    def authorized(self) -> bool:
        """
        Checks if the stored API key is valid
        :return: True if valid, False if not (for example because it expired)
        """
        data = self.execute_api_call("authorize", "GET", True)
        return data["status"] == "ok"

    def get_current_matchday_matches(self) -> List[Match]:
        """
        Retrieves a list of matches for the current matchday
        :return: The list of matches
        """
        data = self.execute_api_call("match?matchday=-1", "GET", True)
        return [Match.from_json(x) for x in data["data"]["matches"]]

    def place_bets(self, bets: List[Bet]):
        """
        Places a list of bets
        :param bets: The bets to place
        :return: None
        """
        bet_dict = {}
        for bet in bets:
            bet_dict.update(bet.to_dict())
        data = self.execute_api_call("bet", "PUT", True, bet_dict)
        if data["status"] != "ok":
            self.logger.error("Failed to place bets")
        else:
            new = data["data"]["new"]
            updated = data["data"]["updated"]
            self.logger.info(f"Placed bets ({new} new, {updated} updated) "
                             f"(user:{self.username})")

    def execute_api_call(
            self,
            endpoint: str,
            method: str,
            authorization_required: bool = False,
            json_data: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """
        Executes an API call
        :param endpoint: The API endpoint
        :param method: The request method
        :param authorization_required: Whether authoirzation is required
        :param json_data: The JSON data to send
        :return: The response JSON
        """
        if authorization_required and endpoint != "authorize":
            logged_in = self.login()
            if not logged_in:
                return {"status": "error"}

        extras = {}
        if authorization_required:
            extras["headers"] = self.auth_headers
        if json_data is not None:
            extras["json"] = json_data

        resp = requests.request(method, self.url + endpoint, **extras)
        return json.loads(resp.text)



#!/usr/bin/env python3
"""LICENSE
Copyright 2017 Hermann Krumrey <hermann@krumreyh.com>

This file is part of bundesliga-tippspiel.

bundesliga-tippspiel is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

bundesliga-tippspiel is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with bundesliga-tippspiel.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import time
import logging
from betbot.api.ApiConnection import ApiConnection
from betbot.prediction import predictors


def main(predictor_name: str, username: str, password: str, url: str):
    """
    The main function of the betbot
    Periodically predicts results and places bets accordingly
    :param predictor_name: The name of the predictor to use
    :param username: The username for bundesliga-tippspiel
    :param password: The password for bundesliga-tippspiel
    :param url: The base URL for the bundesliga-tippspiel instance
    :return: None
    """
    logging.info(f"Starting betbot for predictor {predictor_name} "
                 f"and user {username}@{url}")

    predictor_map = {
        x.name(): x for x in predictors
    }
    predictor = predictor_map[predictor_name]()
    api = ApiConnection(username, password, url)

    while True:
        matches = api.get_current_matchday_matches()
        bets = predictor.predict(matches)
        api.place_bets(bets)
        time.sleep(60 * 60)



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of betbot.

betbot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

betbot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with betbot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Dict


class Bet:
    """
    Class that encapsulates Bet information
    """

    def __init__(self, match_id: int, home_score: int, away_score: int):
        """
        Initializes the Bet
        :param match_id: The ID of the associated match
        :param home_score: The score bet on the home team
        :param away_score: The score bet on the away team
        """
        self.match_id = match_id
        self.home_score = home_score
        self.away_score = away_score

    def to_dict(self) -> Dict[str, int]:
        """
        :return: A dictionary that canbe used to place the bet using the API
        """
        return {
            f"{self.match_id}-home": self.home_score,
            f"{self.match_id}-away": self.away_score
        }



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of betbot.

betbot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

betbot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with betbot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import random
from typing import List
from betbot.api.Bet import Bet
from betbot.api.Match import Match
from betbot.prediction.Predictor import Predictor


class DrawPredictor(Predictor):
    """
    Class that always predicts 0:0, 1:1, 2:2 or 3:3
    """

    @classmethod
    def name(cls) -> str:
        """
        :return: The name of the predictor
        """
        return "draw"

    def predict(self, matches: List[Match]) -> List[Bet]:
        """
        Performs the prediction
        :param matches: The matches to predict
        :return: The predictions as Bet objects
        """
        bets = []
        for match in matches:
            goals = random.randint(0, 3)
            bets.append(Bet(match.id, goals, goals))
        return bets
